import logging
import subprocess
import requests
from helpers import synchronized, URL, Dict_Diff
from job import Job
from plugin import Plugin


class Plugin_Manager(object):

    REFRESH_JOB_NAME = 'refresh_plugins'

    def __init__(self, scheduler, context, cache_manager, response_handler):
        self.__scheduler = scheduler
        self.__context = context
        self.__plugins = {}
        self.__cache_manager = cache_manager
        self.__response_handler = response_handler

        self.__refresh()
        self.__schedule_plugin_refresh()

    def __schedule_plugin_refresh(self):
        refreshJob = Job(self.REFRESH_JOB_NAME, self.__refresh, None, 10)
        self.__scheduler.add_job(refreshJob)

    @synchronized
    def __refresh(self):
        logging.info('refreshing plugins')
        latest = self.__get_latest_plugins()
        self.__update_plugin_list(latest)

    def __get_latest_plugins(self):
        plugin_list = self.__get_plugin_list()
        return self.__get_latest_plugins_from_list(plugin_list)

    def __get_latest_plugins_from_list(self, plugin_list):
        latest_plugins = {}
        for plugin in plugin_list:
            latest_plugins[plugin['name']] = self.__get_plugin_detail(plugin)

        return latest_plugins

    def __get_plugin_list(self):
        result = requests.get(self.__context.plugin_api)
        return result.json()

    def __get_plugin_detail(self, plugin):
        detail = requests.get(URL.join(self.__context.plugin_api, plugin['name']))
        return self.__create_plugin(detail.json())

    def __create_plugin(self, details):
        plugin_name = details['name']
        plugin_location = self.__cache_manager.get_cache_location(plugin_name)
        return Plugin(plugin_name, details['description'],
                      details['interval'], 'alfresco', details['file_hash'],
                      plugin_location)

    def __update_plugin_list(self, latest):
        if len(latest):
            diff = Dict_Diff(self.__plugins, latest)
            self.__add_plugins(diff.added_values())
            self.__remove_plugins(diff.removed_values())
            self.__update_plugins(diff.changed_values())
            self.__plugins = latest

    def __add_plugins(self, plugins):
        for plugin in plugins:
            self.__update_plugin(plugin)

    def __remove_plugins(self, plugins):
        for plugin in plugins:
            self.__remove_plugin(plugin)

    def __remove_plugin(self, plugin):
        self.__scheduler.remove_job(plugin.name)

    def __update_plugins(self, plugins):
        for plugin in plugins:
            self.__update_plugin(plugin)

    def __update_plugin(self, plugin):
        self.__remove_plugin(plugin)
        self.__add_plugin(plugin)

    def __add_plugin(self, plugin):
        self.__cache_manager.load(plugin)
        self.__schedule_plugin(plugin)

    def __schedule_plugin(self, plugin):
        job = self.__create_job(plugin)
        self.__scheduler.add_job(job)

    def __create_job(self, plugin):
        return Job(plugin.name, self.__execute, plugin, plugin.interval)

    def __execute(self, plugin):
        logging.info("exec plugin %s", plugin)
        if self.__cache_manager.is_plugin_cached(plugin):
            self.__run_plugin(plugin)

    def __run_plugin(self, plugin):
        try:
            response = self.__execute_plugin(plugin)
            self.__response_handler.handle_response(plugin, response)
        except Exception, e:
            logging.exception("error executing plugin %s: %s", plugin.name, e)
            raise

    def __execute_plugin(self, plugin):
        logging.info("exec plugin %s", plugin)
        process = subprocess.Popen(plugin.location,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        response = []
        output, errors = process.communicate()
        return_code = process.returncode
        response.append(return_code)
        response.append(output)
        logging.error("subprocess error: %s output: %s",  errors, output)
        return response